SOUTH, NORTH, EAST, WEST, PICKUP = 0, 1, 2, 3, 4
ACTIONS = [SOUTH, NORTH, EAST, WEST, PICKUP]

STEP_REWARD, APPLE_PICKUP_REWARD, BAD_APPLE_PICKUP_REWARD, STEP_INTO_THORN_REWARD = "step", "good_pickup", "bad_pickup", "step_into_thorn"
REWARD_DICT = {STEP_REWARD: -1, APPLE_PICKUP_REWARD: 20, BAD_APPLE_PICKUP_REWARD: -10, STEP_INTO_THORN_REWARD: -5}

ABOVE = -1
UNDER = 1
COLLECTED = 1
NOT_COLLECTED = 0

DETERMINISTIC_PROB = 1.0
STOCHASTIC_PROB = 0.60
STOCHASTIC_PROB_THORNY = 1 - STOCHASTIC_PROB
