TAXIS_LOC_IDX = 0
FUELS_IDX = 1
PASS_START_LOC_IDX = 2
PASS_DEST_IDX = 3
PASS_STATUS_IDX = 4

TAXI = "taxi"
TAXI_EXAMPLE = "taxi_example"
SINGLE_TAXI_EXAMPLE = "single_taxi_example"
SINGLE_FROZEN_EXAMPLE = "single_frozens_example"

LUNAR_LANDER = "lunar_lander"
DIM_REDUCTION_IDX = PASS_STATUS_IDX
ENV_NAME = "taxi"

TAXI_LOC_X = "TX_"
TAXI_LOC_Y = "TY_"
FUEL = "F_"
PASS_LOC = "PL_"
PASS_DEST = "PD_"
PASS_LOC_X = "PX_"
PASS_LOC_Y = "PY_"
PASS_DEST_X = "DX_"
PASS_DEST_Y = "DY_"
PASS_STATUS = "S_"
WALLS = "W_"
REWARD = "R_"
STATE_VISIBILITY_TRANSFORM = "SVT"
ALL_OUTCOME_DETERMINIZATION = "AOD_"
MOST_LIKELY_OUTCOME = "MLO_"
PRECONDITION_RELAXATION = "PR_"

X_COORDINATE = "XC_"
Y_COORDINATE = "YC_"
X_VELOCITY = "XV_"
Y_VELOCITY = "YV_"
ORIENTATION = "OR_"
ANGULAR_VELOCITY = "AV_"



